/* 	llg1_6.c - part of Spinsim_GSL Project
	This file only contains the function llg1_6()
	It is used by spinsim.c as the differential equation to solve for the magnetization
	of a 3-layer structure, having one pinned and two free layers 
	Written and Engineered 2010-01-16 S. Moyerman
	University of California San Diego, ECE Department, Fullerton Lab
*/

/* llg1_6() is a function that is the "differential equation" for the ODE solver, when solving for a system with three layers
It takes parameters as defined by GSL odeiv system (http://www.gnu.org/software/gsl/manual/html_node/Defining-the-ODE-System.html)
t = the time position to use
y[] = the values of y at which our equation is being computed
dydt[] = the array in which to store the output
params = a pointer to a memory space of 19 pointers that point to the following variables:
	{p, Hz, j, Beta_div_P_and_j, Alpha, Alpha_bar, H_K1, H_K2, thetaoffset, ND, NR, Ms, MRs, m_x, m_y, m_z, Gamma_bar, k, i, HP_K1}
The function will return the values of dy/dt for each of the three dimensions STORED INTO dydt[]
The actual return value is GSL_SUCCESS for a valid execution */
int llg1_6(double t, const double y[], double dydt[], void * params) {
//	First, extract all values from params parameter
//	Ivan: We make a poor non-portable assumption that pointers to all types have same length
//	But it's probably good enough for most architectures
	double ** param_list = (double **) params;
	double * p = param_list[0];
	double * Hz = param_list[1];
	double * j = param_list[2];
	double Beta_div_P_and_j = *param_list[3];
	double Alpha = *param_list[4];
	double Alpha_bar = *param_list[5];
	double H_K1 = *param_list[6];
	double H_K2 = *param_list[7];
	double thetaoffset = *param_list[8];
	gsl_matrix* ND = (gsl_matrix*) param_list[9];
	gsl_matrix* NR = (gsl_matrix*) param_list[10];
	double Ms = *param_list[11];
	double MRs = *param_list[12];
	double mx_0 = *param_list[13];
	double my_0 = *param_list[14];
	double mz_0 = *param_list[15];
	double Gamma_bar = *param_list[16];
	int k = *((int*)param_list[17]);
	int i = *((int*)param_list[18]);
	double HP_K1 = *param_list[19]; //anisotropy for the pinned layer
	int j_timedep_funs = *((int*)param_list[20]);
	int hz_timedep_funs = *((int*)param_list[21]);
	
//	This is the current density we'll use for this iterations - pre-calculate it so we don't have to recalculate each time we use it
	double this_iterations_j = TimeDepFun(CALC_CURRENT, t, j[i], j_timedep_funs);

//	Now we'll define and calculate variable's we'll actually use
	double Beta = Beta_div_P_and_j*polarisation(y[0],y[1],y[2],y[3],y[4],y[5])*this_iterations_j;
	double Beta_bar1 = Alpha*Beta/( 1 + pow(Alpha,2) );
	double Beta_bar2 = Beta/(1+pow(Alpha,2));

	double Betap = -Beta_div_P_and_j*polarisation(y[3],y[4],y[5],y[0],y[1],y[2])*this_iterations_j;
	double Beta_bar1p = Alpha*Betap/( 1 + pow(Alpha,2) );
	double Beta_bar2p = Betap/(1+pow(Alpha,2));

//	Re-define y as y_mat, a matrix with the values as different columns
	gsl_matrix* y_mat = gsl_matrix_alloc(3,1);
	gsl_matrix_set(y_mat, 0, 0, y[0]);
	gsl_matrix_set(y_mat, 1, 0, y[1]);
	gsl_matrix_set(y_mat, 2, 0, y[2]);
//	Re-define p as p_mat, a matrix with the values as different columns
	gsl_matrix* p_mat = gsl_matrix_alloc(3,1);
	gsl_matrix_set(p_mat, 0, 0, y[3]);
	gsl_matrix_set(p_mat, 1, 0, y[4]);
	gsl_matrix_set(p_mat, 2, 0, y[5]);

//	  Free Layer
//	Define H, the applied field, as a gsl_matrix, remember that it's (rows, columns)
	gsl_matrix* H = gsl_matrix_alloc(3,1);
	gsl_matrix_set(H, 0, 0, 0.0);
	gsl_matrix_set(H, 1, 0, 0.0);
	gsl_matrix_set(H, 2, 0, TimeDepFun(CALC_FIELD, t, Hz[k], hz_timedep_funs));
//	Anisotropic field
	gsl_matrix* Hani = gsl_matrix_alloc(3,1);
	gsl_matrix_set(Hani, 0, 0, H_K1*y[0]*sin(thetaoffset));
	gsl_matrix_set(Hani, 1, 0, 0.0);
	gsl_matrix_set(Hani, 2, 0, H_K1*y[2]*cos(thetaoffset)+H_K2*y[2]*( pow(y[1],2) + pow(y[2],2) ) );
//	Self-demag field "Hani"
	gsl_matrix* HD = gsl_matrix_alloc(3,1);
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, ND, y_mat, 0, HD);
	gsl_matrix_scale(HD, (-1.0 * Ms));
//	Mutual demag field "HR"
	gsl_matrix* HR = gsl_matrix_alloc(3,1);
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, NR, p_mat, 0, HR);
	gsl_matrix_scale(HR, (-1.0 * MRs));
//	Effective field is the sum of the anisotropy field (Hani), the self (HD)
//	and mutual (HR) demag. fields, and the external field (H)
	gsl_matrix* Heff = gsl_matrix_calloc(3,1); //calloc sets initial values to 0
	gsl_matrix_add(Heff, H);
	gsl_matrix_add(Heff, Hani);
	gsl_matrix_add(Heff, HR);
	gsl_matrix_add(Heff, HD);

//	  Pinned Layer
	gsl_matrix* HPani = gsl_matrix_alloc(3,1);
	gsl_matrix_set(HPani, 0, 0, 0.0);
	gsl_matrix_set(HPani, 1, 0, 0.0);
	gsl_matrix_set(HPani, 2, 0, HP_K1*y[5]);
	
	gsl_matrix* HPD = gsl_matrix_alloc(3,1);
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, ND, p_mat, 0, HPD);
	gsl_matrix_scale(HPD, (-1.0 * MRs));
	
	gsl_matrix* HPR = gsl_matrix_alloc(3,1);
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, NR, y_mat, 0, HPR);
	gsl_matrix_scale(HPR, (-1.0 * Ms));
	
	gsl_matrix* HPeff = gsl_matrix_alloc(3,1); //calloc sets initial values toasdfasdfasdfasdf																		 
	gsl_matrix_set_zero(HPeff);
	gsl_matrix_add(HPeff, H);
	gsl_matrix_add(HPeff, HPani);
	gsl_matrix_add(HPeff, HPR);
	gsl_matrix_add(HPeff, HPD);

//	In fact, one has to integrate three differential equations,
//	one for each component of m

//The original matlab code
/*dy(1) = -Gamma_bar*(y(2)*Heff(3)-y(3)*Heff(2))-Alpha_bar*((y(1)*y(1)-1)*Heff(1)+...
		 y(1)*y(2)*Heff(2)+y(1)*y(3)*Heff(3))+Beta_bar1*(y(2)*y(6) - y(3)*y(5))...
		   -Beta_bar2*((y(1)*y(1)-1)*y(4)+y(1)*y(2)*y(5)+y(1)*y(3)*y(6));

	 dy(2) = -Gamma_bar*(y(3)*Heff(1)-y(1)*Heff(3))-Alpha_bar*((y(2)*y(2)-1)*Heff(2)+...
		 y(1)*y(2)*Heff(1)+y(2)*y(3)*Heff(3))+Beta_bar1*(y(3)*y(4)-y(1)*y(6))...
		   -Beta_bar2*((y(2)*y(2)-1)*y(5)+y(1)*y(2)*y(4)+y(2)*y(3)*y(6));

	 dy(3) = -Gamma_bar*(y(1)*Heff(2)-y(2)*Heff(1))-Alpha_bar*((y(3)*y(3)-1)*Heff(3)...
		 +y(1)*y(3)*Heff(1)+y(2)*y(3)*Heff(2))+Beta_bar1*(y(1)*y(5)-y(2)*y(4))....
		   -Beta_bar2*((y(3)*y(3)-1)*y(6)+y(1)*y(3)*y(4)+y(2)*y(3)*y(5));
	
	 dy(4) = -Gamma_bar*(y(5)*HPeff(3)-y(6)*HPeff(2))-Alpha_bar*((y(4)*y(4)-1)*HPeff(1)+...
		 y(4)*y(5)*HPeff(2)+y(4)*y(6)*HPeff(3))+Beta_bar1p*(y(5)*y(3) - y(6)*y(2))...
		   -Beta_bar2p*((y(4)*y(4)-1)*y(1)+y(4)*y(5)*y(2)+y(4)*y(6)*y(3));

	 dy(5) = -Gamma_bar*(y(6)*HPeff(1)-y(4)*HPeff(3))-Alpha_bar*((y(5)*y(5)-1)*HPeff(2)+...
		 y(4)*y(5)*HPeff(1)+y(5)*y(6)*HPeff(3))+Beta_bar1p*(y(6)*y(1)-y(4)*y(3))...
		   -Beta_bar2p*((y(5)*y(5)-1)*y(2)+y(4)*y(5)*y(1)+y(5)*y(6)*y(3));

	 dy(6) = -Gamma_bar*(y(4)*HPeff(2)-y(5)*HPeff(1))-Alpha_bar*((y(6)*y(6)-1)*HPeff(3)...
		 +y(4)*y(6)*HPeff(1)+y(5)*y(6)*HPeff(2))+Beta_bar1p*(y(4)*y(2)-y(5)*y(1))....
		   -Beta_bar2p*((y(6)*y(6)-1)*y(3)+y(4)*y(6)*y(1)+y(5)*y(6)*y(2));*/

	// Free Layer Dynamics
	dydt[0] = -1.0 * Gamma_bar * (y[1]*gsl_matrix_get(Heff, 2, 0)-y[2]*gsl_matrix_get(Heff, 1, 0)) - Alpha_bar*((y[0]*y[0]-1)*gsl_matrix_get(Heff, 0, 0)+y[0]*y[1]*gsl_matrix_get(Heff, 1, 0)+y[0]*y[2]*gsl_matrix_get(Heff, 2, 0)) + Beta_bar1 * (y[1]*y[5] - y[2]*y[4]) - Beta_bar2 * ((y[0]*y[0]-1)*y[3]+y[0]*y[1]*y[4]+y[0]*y[2]*y[5]);
	dydt[1] = -1.0 * Gamma_bar*(y[2]*gsl_matrix_get(Heff, 0, 0)-y[0]*gsl_matrix_get(Heff, 2, 0))-Alpha_bar*((y[1]*y[1]-1)*gsl_matrix_get(Heff, 1, 0)+y[0]*y[1]*gsl_matrix_get(Heff, 0, 0)+y[1]*y[2]*gsl_matrix_get(Heff, 2, 0))+Beta_bar1*(y[2]*y[3]-y[0]*y[5]) - Beta_bar2*((y[1]*y[1]-1)*y[4]+y[0]*y[1]*y[3]+y[1]*y[2]*y[5]);
	dydt[2] = -1.0 * Gamma_bar*(y[0]*gsl_matrix_get(Heff, 1, 0)-y[1]*gsl_matrix_get(Heff, 0, 0))-Alpha_bar*((y[2]*y[2]-1)*gsl_matrix_get(Heff, 2, 0)+y[0]*y[2]*gsl_matrix_get(Heff, 0, 0)+y[1]*y[2]*gsl_matrix_get(Heff, 1, 0))+Beta_bar1*(y[0]*y[4]-y[1]*y[3])-Beta_bar2*((y[2]*y[2]-1)*y[5]+y[0]*y[2]*y[3]+y[1]*y[2]*y[4]);
	
	// Ref Layer Dynamics
	dydt[3] = -1.0 * Gamma_bar*(y[4]*gsl_matrix_get(HPeff, 2, 0)-y[5]*gsl_matrix_get(HPeff, 1, 0)) - Alpha_bar*((y[3]*y[3]-1)*gsl_matrix_get(HPeff, 0, 0)+y[3]*y[4]*gsl_matrix_get(HPeff, 1, 0)+y[3]*y[5]*gsl_matrix_get(HPeff, 2, 0))+Beta_bar1p*(y[4]*y[2]-y[5]*y[1])-Beta_bar2p*((y[3]*y[3]-1)*y[0]+y[3]*y[4]*y[1]+y[3]*y[5]*y[2]);
	dydt[4] = -1.0*Gamma_bar*(y[5]*gsl_matrix_get(HPeff, 0, 0)-y[3]*gsl_matrix_get(HPeff, 2, 0))-Alpha_bar*((y[4]*y[4]-1)*gsl_matrix_get(HPeff, 1, 0)+y[3]*y[4]*gsl_matrix_get(HPeff, 0, 0)+y[4]*y[5]*gsl_matrix_get(HPeff, 2, 0))+Beta_bar1p*(y[5]*y[0]-y[3]*y[2])-Beta_bar2p*((y[4]*y[4]-1)*y[1]+y[3]*y[4]*y[0]+y[4]*y[5]*y[2]);
	dydt[5] = -1.0*Gamma_bar*(y[3]*gsl_matrix_get(HPeff, 1, 0)-y[4]*gsl_matrix_get(HPeff, 0, 0))-Alpha_bar*((y[5]*y[5]-1)*gsl_matrix_get(HPeff, 2, 0)+y[3]*y[5]*gsl_matrix_get(HPeff, 0, 0)+y[4]*y[5]*gsl_matrix_get(HPeff, 1, 0))+Beta_bar1p*(y[3]*y[1]-y[4]*y[0])-Beta_bar2p*((y[5]*y[5]-1)*y[2]+y[3]*y[5]*y[0]+y[4]*y[5]*y[1]);

	//free things we've used, mostly all of the matrices
	gsl_matrix_free(y_mat);
	gsl_matrix_free(p_mat);
	
	//free free layer
	gsl_matrix_free(Heff);
	gsl_matrix_free(HR);	
	gsl_matrix_free(Hani);
	gsl_matrix_free(HD);
	gsl_matrix_free(H);
	
	//free pinned layer
	gsl_matrix_free(HPani);
	gsl_matrix_free(HPeff);
	gsl_matrix_free(HPR);
	gsl_matrix_free(HPD);

	return(GSL_SUCCESS);
}

