/* 	spin_helpers.c
	This file provides various helper functions for the spinsim.c top-level file.
	Note: This was written and tested on Linux 2.6.18 x86-64 with GSL version 1.12, it is recommended to use
	the same versions to prevent compatibility issues.
	Note: If compiling using gcc, be sure to use the '-lgsl -lgslcblas -lm' options on the command line.
	Written I. Yulaev 2010-01-04 (iyulaev@ucsd.edu)
	Engineered S. Moyerman (smoyerman@ucsd.edu)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

//Includes from GSL library
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>

//math.h should include PI but whatever
#ifndef PI
#define PI (3.141592653589793)
#endif

/* Returns the Newell_f output given parameters a,b,c
TODO: Add some relevant description.
Translated from Matlab "Newell_f.m" 2009-12-31 */
double newell_f(double a, double b, double c) {
	if(a == 0) {
		a = pow(10.0,(-20.0));
	}
	if(b == 0) {
		b = pow(10.0,(-20.0));
	}
	if(c == 0) {
		c = pow(10.0,(-20.0));
	}
	
//	calculate Newellf
	double output;
	
	double a_2 = pow(a,2);
	double b_2 = pow(b,2);
	double c_2 = pow(c,2);
	
	output = 0.5*b*(c_2 - a_2) * asinh( b/sqrt(a_2+c_2) ) + 0.5 * c * (b_2-a_2) * asinh( c/sqrt(a_2+b_2) ) - a * b * c * atan((b*c)/(a*sqrt(a_2+b_2+c_2)))+(2*a_2-b_2-c_2)/6*sqrt( a_2+b_2+c_2 );

	return(output);
}

/* Calculates f2(), called from f1()
Used by demagtensor() to put together a tensor
Translated from Matlab "f2.m" 2010-01-01 */
double f2(double a, double b, double c) {
	double returned = newell_f(a,b,c)-newell_f(a,0,c)-newell_f(a,b,0)+newell_f(a,0,0);
	return(returned);
}


/* Calculates f1(), called by f(),
Used by demagtensor() to put together a tensor
Translated from Matlab "f1.m" 2010-01-01 */
double f1(double x, double y, double z, double dx, double dy, double dz) {
	double returned = f2(x,y,z)-f2(x,y-dy,z)-f2(x,y,z-dz)+f2(x,y-dy,z-dz);
	return(returned);
}

/* Calculates f(), called by Nii(),
Used by demagtensor() to put together a tensor
Translated from Matlab "f.m" 2010-01-01 */
double f(double x, double y, double z, double dx, double dy, double dz) {
	double returned = f1(x,y+dy,z+dz,dx,dy,dz)-f1(x,y,z+dz,dx,dy,dz)-f1(x,y+dy,z,dx,dy,dz)+f1(x,y,z,dx,dy,dz);
	return(returned);
}

/* Calculates nii(), called by demagtensor(),
Used by demagtensor() to put together a tensor
Translated from Matlab "Nii.m" 2010-01-01 */
double nii(double x, double y, double z, double dx, double dy, double dz) {
	double returned = (1/(4*PI*dx*dy*dz))*(2*f(x,y,z,dx,dy,dz)-f(x+dx,y,z,dx,dy,dz)-f(x-dx,y,z,dx,dy,dz));
	return(returned);
}

/* Returns the Newell_g output given parameters a,b,c
TODO: Add some relevant description.
Translated from Matlab "Newell_g.m" 2010-01-01 */
double newell_g(double a, double b, double c) {
	if(a == 0) {
		a = pow(10.0,(-20.0));
	}
	if(b == 0) {
		b = pow(10.0,(-20.0));
	}
	if(c == 0) {
		c = pow(10.0,(-20.0));
	}
	
//	calculate Newellg
	double output;
	
	double a_2 = pow(a,2);
	double b_2 = pow(b,2);
	double c_2 = pow(c,2);
	double c_3 = pow(c,3);
	
	output = a*b*c*asinh(c/sqrt(a_2+b_2))+b/6*(3*c_2-b_2)*asinh(a/sqrt(b_2+c_2))+a/6*(3*c_2-a_2)*asinh(b/sqrt(a_2+c_2))-c_3/6*atan((a*b)/(c*sqrt(a_2+b_2+c_2)))-c*b_2/2*atan((a*c)/(b*sqrt(a_2+b_2+c_2)))-c*a_2/2*atan((b*c)/(a*sqrt(a_2+b_2+c_2)))-a*b*sqrt(a_2+b_2+c_2)/3;

	return(output);
}

/* Calculates g2(), called from g1()
Used by demagtensor() to put together a tensor
Translated from Matlab "g2.m" 2010-01-01 */
double g2(double a, double b, double c) {
	double returned = newell_g(a,b,c)-newell_g(a,b,0.0);
	return(returned);
}


/* Calculates g1(), called by g(),
Used by demagtensor() to put together a tensor
Translated from Matlab "g1.m" 2010-01-01 */
double g1(double x, double y, double z, double dx, double dy, double dz) {
	double returned = g2(x+dx,y,z+dz)-g2(x+dx,y,z)-g2(x,y,z+dz)+g2(x,y,z);
	return(returned);
}

/* Calculates g(), called by Nij(),
Used by demagtensor() to put together a tensor
Translated from Matlab "g.m" 2010-01-01 */
double g(double x, double y, double z, double dx, double dy, double dz) {
	double returned = g1(x,y,z,dx,dy,dz)-g1(x,y-dy,z,dx,dy,dz)-g1(x,y,z-dz,dx,dy,dz)+g1(x,y-dy,z-dz,dx,dy,dz);
	return(returned);
}

/* Calculates nij(), called by demagtensor(),
Used by demagtensor() to put together a tensor (TODO: describe better)
Translated from Matlab "Nij.m" 2010-01-01 */
double nij(double x, double y, double z, double dx, double dy, double dz) {
	double returned = (1/(4*PI*dx*dy*dz))*(g(x,y,z,dx,dy,dz)-g(x-dx,y,z,dx,dy,dz)-g(x,y+dy,z,dx,dy,dz)+g(x-dx,y+dy,z,dx,dy,dz));
	return(returned);
}

/* This function puts together a demagnetizing tensor given the parameters
The return value is status (0 means good)
The matrix is put into the pointer '* output_tensor'
Translated from Matlab "demagtensor.m" 2010-01-01 */
int demagtensor(double x, double y, double z, double dx, double dy, double dz, gsl_matrix* output_tensor) {
//	The demagnetizing tensor is stored as a matrix of values that looks like:
//	( Nxx  Nxy  Nxz )
//	( Nyx  Nyy  Nyz )
//	( Nzx  Nzy  Nzz )
//	We use gsl_matrix_set to set the values in the matrix (already allocated for us) accordingly
	gsl_matrix_set(output_tensor, 0, 0, nii(x,y,z,dx,dy,dz));
	gsl_matrix_set(output_tensor, 0, 1, nij(x,y,z,dx,dy,dz));
	gsl_matrix_set(output_tensor, 0, 2, nij(x,z,y,dx,dz,dy));
	gsl_matrix_set(output_tensor, 1, 0, nij(y,x,z,dy,dx,dz));
	gsl_matrix_set(output_tensor, 1, 1, nii(y,x,z,dy,dx,dz));
	gsl_matrix_set(output_tensor, 1, 2, nij(y,z,x,dy,dz,dx));
	gsl_matrix_set(output_tensor, 2, 0, nij(z,x,y,dz,dx,dy));
	gsl_matrix_set(output_tensor, 2, 1, nij(z,y,x,dz,dy,dx));
	gsl_matrix_set(output_tensor, 2, 2, nii(z,y,x,dz,dy,dx));

	return(0);
}

/*Polarisation computes the polarisation of something or other (TODO)
Called by the differential equation function
Translated from Matlab "polarisation.m" 2010-01-02 */
double polarisation(double m1_x, double m1_y, double m1_z, double m2_x, double m2_y, double m2_z) {
	double B0=1;
	double B1=0.5;
	double Qp=1;
	double Qn=-0.2; //-0.1;

	double cosTheta = m1_x*m2_x+m1_y*m2_y+m1_z*m2_z;
	double returned = Qp/(B0+B1*cosTheta)+Qn/(B0-B1*cosTheta);
	return(returned);
}

/*This function creates a double linspace and sticks it into output
The linspace will go from (-1 * magnitude) to magnitude with 'size' positions
TODO: Allow this to set the end points arbitrarily
Written Ivan Yulaev 2010-01-02
*/
void make_linspace(double min, double max, int size, double * output) {
	int count;
	
	double magnitude = (max - min)/2;
	
	for(count = 0; count < size; count++) {
		output[count] = (-1.0 * magnitude) + (2.0 * ( ((double)count) / ((double)(size-1)) ) * magnitude);
	}
}

/*main here for testing purposes only
int main() {
	int count;
	gsl_matrix* test_m = gsl_matrix_alloc(3,3);

	demagtensor(0.3, 0.5, 0.7, 0.03, 0.05, 7e-2, test_m);

	for(count = 0; count < 3; count++) {
		printf("%f\t%f\t%f\n", gsl_matrix_get(test_m, count, 0), gsl_matrix_get(test_m, count, 1), gsl_matrix_get(test_m, count, 2));
	}
	
		double * blah = NULL;
	
	printf("The size of a double pointer is %d\n", sizeof(blah));
	printf("The size of a int pointer is %d\n", sizeof(int *));
	printf("The size of a gsl_matrix pointer is %d\n", sizeof(gsl_matrix*));
	
	gsl_matrix* y_mat = gsl_matrix_alloc(3,1);
	gsl_matrix_set(y_mat, 0, 0, 2);
	gsl_matrix_set(y_mat, 1, 0, 4);
	gsl_matrix_set(y_mat, 2, 0, 6);
	
	gsl_matrix* result = gsl_matrix_alloc(3,1);
	
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, test_m, y_mat, 0, result);
	
	printf("Result looks like:\n%f\n%f\n%f\n", gsl_matrix_get(result, 0, 0), gsl_matrix_get(result, 1, 0), gsl_matrix_get(result, 2, 0));

	return(0);
}*/
