/* 	spinsim.c
	This file is the top level code for the spinsim simulator. It allows simulation of the behavior of a free layer
	across a grid of externally applied current/H-field conditions. The function to call is Hsweep()
	For greater detail, see documentation PDF for this project.
	Note: This was written and tested on Linux 2.6.18 x86-64 with GSL version 1.12, it is recommended to use
	the same versions to prevent compatibility issues.
	Note: If compiling using gcc, be sure to use the '-pthread -lgsl -lgslcblas -lm' options on the command line.
	Written I. Yulaev 2010-01-xx (iyulaev@ucsd.edu)
	Engineered S. Moyerman (smoyerman@ucsd.edu)
*/

//Defines for general program parameters
#define NOISINESS 1		//How talkative? 1 is "as loud as possible", a large number makes it quiet (like 10000)

//Do we want to do two layers, or three layers
//(doesn't work yet)
#define USE_THREE_LAYERS 1
//How many dimensions in the integrator?
#define NUMBER_OF_DIMENSIONS 6

//Defines for integrator
#define ABS_PREC 1e-6			//Absolute precision for integrator to aim for
#define REL_PREC 1e-6			//Relative precision for integrator to aim for
#define INIT_STEP_SZ 100000.0	//Define the proportion that the initial step will be of t_f, i.e. it will be t_f/INIT_STEP_SZ

//Command-line argument stuff
#define NUMARGS 19

//Defines for multi-threading
//#define NUM_THREADS 1	//define # of threads to use

#include <stdio.h>
#include <math.h>
#include <time.h>

//Includes from GSL Library
#include <gsl/gsl_errno.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_odeiv.h>

//Includes for PThreads API
#include <sys/types.h>
#include <pthread.h>

//Include helper functions like demagtensor and polarisation
#include "spin_helpers.c"
//Include the functions for calculating time dependent field and current
#include "timedepfuns.c"
//Include the differential equation functions
#include "llg1.c"
#include "llg1_6.c"

/* Hsweep() calculates the magnetization of a free layer across a 2D plane of external field & applied current conditions
Parameters:
(1) n = the number of current values to evaluate at
(2) nfield = the number of applied h-field values to evaluate at
(3) percK2 = the percentage that K2 is of K1
(4) thetaoff = the offset of the applied h-field vs the z-axis of the structure
(5) Hz = a 1D-array with size nfield containing the h-field conditions applied to the structure
(6) magsweepup = a 2D array of size (n, nfield) for the magnetization on upsweeping calculation
(7) magsweepdown = a 2D array of size (n, nfield) for the magnetization on downsweeping calculation
(8) j = a 1D-array with size n of the applied current conditions
(9-11) dx/dy/dz = parameters for defining the shape of the demagtensor
(12) Alpha = Gilbert's damping factor, dimensionless
(13) Ms = Free layer saturation magnetization, A/m
(14) Mrs = Pinned layer saturation magnetization, A/m
(15) K1 = Free layer anisotropy constant, J/m^3
(16) hz_min = minimum magnitude of magnetic field to apply
(17) hz_max = maximum magnitude of magnetic field to apply
(18) j_min = minimum current density to apply
(19) j_max = maximum current density to apply
(20) t_f = integration fime
(21) thread_no = the thread number this is. 0 = first thread
(22) numthreads = total # of threads to run
(23) j_timedep_funs = the bitwise-OR of the time dependent functions to use for current density
(24) hz_timedep_funs = the bitwise-OR of the time dependent functions to use for external field
*/
int Hsweep(int n, int nfield, double percK2, double thetaoff, double * Hz, double * magsweepup, double * magsweepdown, double * j, double dx, double dy, double dz, double Alpha, double Ms, double MRs, double K1, double hz_min, double hz_max, double j_min, double j_max, double t_f, int thread_no, int numthreads, unsigned int j_timedep_funs, unsigned int hz_timedep_funs) {
//	save start and stop times to benchmark the program speed
	time_t start, stop, printed_last, temp;
	time(&start);

//	Initial conditions in time and magnetization
	double mz_0 = .9; 
	double mx_0 = 0;
	double my_0 = .1;
//	Magnitude of the magnetization vector
	double m = sqrt( pow((mx_0),2) + pow((my_0),2) + pow((mz_0),2) );
//	Make the magnetization vector unit sized
	mx_0 = mx_0 / m; 
	my_0 = my_0 / m;
	mz_0 = mz_0 / m; 

//	Integration times
	double t_0 = 0;

//	Mutual Demagnetization Tensor
	double X = 0;
	double Y = 0;
	double Z = 7;
	gsl_matrix * NR = gsl_matrix_alloc(3,3);
	demagtensor(X,Y,Z,dx,dy,dz,NR);
//	Note: For infinite lateral size use NR_xx = 0; NR_yy = 0; NR_zz = 0;

//	Self Demagnetization Matrix
	X = 0; Y = 0; Z = 0;
	gsl_matrix * ND = gsl_matrix_alloc(3,3);
	demagtensor(X,Y,Z,dx,dy,dz,ND);
//	Note: For infinite lateral size use ND_xx = 0; ND_yy = 0; ND_zz = 1;

//	Define the normal LLG parameters for the problem
	double Gamma = 2.21e5; 			//Gyromagnetic ratio, 2.2127616 x 1e5 s^(-1).A^(-1).m
	double d = 3e-9;               //Free layer thickness, m
	double K2 = (percK2/100)*K1;   //Free layer anisotropy constant
	double mu0 = 4*PI*1e-7; 		//Magnetic permeability of vaccuum, H/m
	double e = 1.602e-19; 			//Electron charge, -1.602176 x 1e-19 C
	double hbar = 1.054e-34; 		//Reduced Planck constant, 1.054571 x 1e-34 J.s

//	Other constants
	double px = 0; 
	double py = 0; 
	double pz = 1;		 //Pinned layer initial state
	double p[3] = {0, 0, 1};
		
//	Slonczewski's spin transfert torque coefficient Beta=P*j*Beta_div_P_and_j
	double Beta_div_P_and_j = (Gamma*hbar)/(2*d*mu0*Ms*e);

//	Some useful parameters
	double H_K1 = 2*K1/mu0/Ms;                   //Free layer anisotropy field, A/m
	double H_K2 = 4*K2/mu0/Ms;                   //Free layer anisotropy field, A/m
	double norm = H_K1 / sqrt( pow(H_K1,2)+pow(H_K2,2) );
	H_K1 = norm*H_K1;
	H_K2 = norm*H_K2;
	//H_K2 = 0; //This doesn't feel right, but it was originally uncommented
	double Gamma_bar = Gamma/(1+pow(Alpha,2) );
	double Alpha_bar = Alpha*Gamma/( 1 + pow(Alpha,2) );
	
//	stuff for the 3rd layer
//	Define params for the reference layer - for reasonable dynamics, make 1.5x K1
	double Kp1 = 3e5*10;          // pinned layer anisotropy
	double HP_K1 = 2*Kp1/mu0/Ms;                        //Pinned layer anisotropy field, A/m

//	Initialize our applied field vector
//	Applied field will go from (-HZ_MAX) to (HZ_MAX) with nfield steps in between
//	TODO: This would probably be better if the space did not have to center about 0
	make_linspace(hz_min,hz_max,nfield,Hz);

//	Current Density Vector
//	Applied current density will go from (-J_MAX) to (J_MAX) with nfield steps in between
	make_linspace(j_min,j_max,n,j); 
	
	int count3;
	for(count3 = 0; count3 < n; count3++) {
		//printf("j[%d] = %e\n", count3, j[count3]);
	}

//	Offset for easy axis of free FM layer
	double thetaoffset = thetaoff*PI/180;
	
//	Loop variables
	int i, k;
	char percent[2] = "%";
//	We'll use this to gauge when the last time we gave a status update was
	time(&printed_last);
	
	for(i = thread_no; i < n; i+=numthreads) {
		time(&temp);
		if((thread_no == 0) && ( (((int)temp) - ((int)printed_last)) >= NOISINESS) ) {
			printf("%.1f%s\n", ((float) i) / ((float) n) * 100.0, percent);
			fflush(stdout);
			time(&printed_last);
		}

//		sweep down
		for(k = (nfield-1); k >= 0; k--) {
//			put together a parameter list...
			double ** param_list = (double **) malloc(22 * sizeof(double *));
			param_list[0] = p;
			param_list[1] = Hz;
			param_list[2] = j;
			param_list[3] = &Beta_div_P_and_j;
			param_list[4] = &Alpha;
			param_list[5] = &Alpha_bar;
			param_list[6] = &H_K1;
			param_list[7] = &H_K2;
			param_list[8] = &thetaoffset;
			param_list[9] = (double *) ND;
			param_list[10] = (double *) NR;
			param_list[11] = &Ms;
			param_list[12] = &MRs;
			param_list[13] = &mx_0;
			param_list[14] = &my_0;
			param_list[15] = &mz_0;
			param_list[16] = &Gamma_bar;
			param_list[17] = (double *) &k; //yes, k is an int, so what
			param_list[18] = (double *) &i; //yes, i is an int, this may not be portable but should be fine for x86(-64)
			param_list[19] = &HP_K1;
			param_list[20] = (double*) &j_timedep_funs;
			param_list[21] = (double*) &hz_timedep_funs;
			
//			set up the ode system (most of this code is lifted from the GSL documentation and just re-written)
//			http://www.gnu.org/software/gsl/manual/html_node/Ordinary-Differential-Equations.html
//			we don't give it a jacobian because we're going to use runge-kutta algorithm which doesn't need it
			
			#ifdef USE_THREE_LAYERS
			gsl_odeiv_system sys = {llg1_6, NULL, NUMBER_OF_DIMENSIONS, param_list};
			#endif
			
			#ifndef USE_THREE_LAYERS
			gsl_odeiv_system sys = {llg1, NULL, NUMBER_OF_DIMENSIONS, param_list};
			#endif
			
//			define step type
//			So far, gsl_odeiv_step_rk4 seems to be the best, but gsl_odeiv_step_rk8pd, gsl_odeiv_step_rk2 isn't bad either. R-K (4,5) appears to suck.
			const gsl_odeiv_step_type * T = gsl_odeiv_step_rk4;
//			define the "stepper", give it the step type and # of dimensions
			gsl_odeiv_step * s = gsl_odeiv_step_alloc (T, NUMBER_OF_DIMENSIONS);
//			Define controller
//			Ivan: Stephanie, please examine this - do we need to use the _y or _yp version? Also parameters to this controller, I just guessed.
//			see http://www.gnu.org/software/gsl/manual/html_node/Adaptive-Step_002dsize-Control.html
			gsl_odeiv_control * c = gsl_odeiv_control_y_new (ABS_PREC, REL_PREC);
//			Allocate the function we're going to use for evolution
			gsl_odeiv_evolve * e = gsl_odeiv_evolve_alloc (NUMBER_OF_DIMENSIONS);
//			Define the starting time for our integration
			double t = t_0;
//			Define the initial step size
			double h = t_f / INIT_STEP_SZ;
//			Make space for the output of the integration
			double y_result[6];
//			Set initial condition
			y_result[0] = mx_0;
			y_result[1] = my_0;
			y_result[2] = mz_0;
			#ifdef USE_THREE_LAYERS
			y_result[3] = px;
			y_result[4] = py;
			y_result[5] = pz; 
			#endif
			
//			Do the actual integration!
			while(t < t_f) {
				int status = gsl_odeiv_evolve_apply (e, c, s, &sys, &t, t_f, &h, y_result);
			
     
				if (status != GSL_SUCCESS) {
					printf("Integrator broke for some reason - help!\n");
					break;
				}
			}
			
//			Store result in magsweepdown
			magsweepdown[i*nfield + k] = y_result[2];

//			Free the things we defined a while ago
			gsl_odeiv_evolve_free (e);
			gsl_odeiv_control_free (c);
			gsl_odeiv_step_free (s);
			free(param_list);
		} //end magsweepdown for loop
		
//		sweep up
		for(k = 0; k < nfield; k++) {
//			put together a parameter list...
			double ** param_list = (double **) malloc(22 * sizeof(double *));
			param_list[0] = p;
			param_list[1] = Hz;
			param_list[2] = j;
			param_list[3] = &Beta_div_P_and_j;
			param_list[4] = &Alpha;
			param_list[5] = &Alpha_bar;
			param_list[6] = &H_K1;
			param_list[7] = &H_K2;
			param_list[8] = &thetaoffset;
			param_list[9] = (double *) ND;
			param_list[10] = (double *) NR;
			param_list[11] = &Ms;
			param_list[12] = &MRs;
			param_list[13] = &mx_0;
			param_list[14] = &my_0;
			param_list[15] = &mz_0;
			param_list[16] = &Gamma_bar;
			param_list[17] = (double *) &k; //yes, k is an int, so what
			param_list[18] = (double *) &i; //yes, i is an int, this may not be portable but should be fine for x86(-64)
			param_list[19] = &HP_K1;
			param_list[20] = (double*) &j_timedep_funs;
			param_list[21] = (double*) &hz_timedep_funs;
			
//			set up the ode system (most of this code is lifted from the GSL documentation and just re-written)
//			http://www.gnu.org/software/gsl/manual/html_node/Ordinary-Differential-Equations.html
//			we don't give it a jacobian because we're going to use runge-kutta algorithm which doesn't need it
			
			#ifdef USE_THREE_LAYERS
			gsl_odeiv_system sys = {llg1_6, NULL, NUMBER_OF_DIMENSIONS, param_list};
			#endif
			
			#ifndef USE_THREE_LAYERS
			gsl_odeiv_system sys = {llg1, NULL, NUMBER_OF_DIMENSIONS, param_list};
			#endif
			
//			define step type
//			So far, gsl_odeiv_step_rk4 seems to be the best, but gsl_odeiv_step_rk8pd isn't bad either. R-K (4,5) appears to suck.
			const gsl_odeiv_step_type * T = gsl_odeiv_step_rk4;
//			define the "stepper", give it the step type and # of dimensions
			gsl_odeiv_step * s = gsl_odeiv_step_alloc (T, NUMBER_OF_DIMENSIONS);
//			Define controller
//			Ivan: Stephanie, please examine this - do we need to use the _y or _yp version? Also parameters to this controller, I just guessed.
//			see http://www.gnu.org/software/gsl/manual/html_node/Adaptive-Step_002dsize-Control.html
			gsl_odeiv_control * c = gsl_odeiv_control_y_new (ABS_PREC, REL_PREC);
//			Allocate the function we're going to use for evolution
			gsl_odeiv_evolve * e = gsl_odeiv_evolve_alloc (NUMBER_OF_DIMENSIONS);
//			Define the starting time for our integration
			double t = t_0;
//			Define the initial step size
			double h = t_f / INIT_STEP_SZ;
//			Make space for the output of the integration
			double y_result[6];
//			Set initial condition
			y_result[0] = mx_0;
			y_result[1] = my_0;
			y_result[2] = (-1.0) * mz_0;
			#ifdef USE_THREE_LAYERS
			y_result[3] = px;
			y_result[4] = py;
			y_result[5] = pz; 
			#endif
			
//			Do the actual integration!
			while(t < t_f) {
				int status = gsl_odeiv_evolve_apply (e, c, s, &sys, &t, t_f, &h, y_result);
     
				if (status != GSL_SUCCESS) {
					printf("Integrator broke for some reason - help!\n");
					break;
				}
			}
			
			//printf("DEBUG: Last results for y[6] were {%.3e, %.3e, %.3e, %.3e, %.3e, %.3e}\n", y_result[0], y_result[1], y_result[2], y_result[3], y_result[4], y_result[5]);
			
			//fprintf(stderr, "Finished integrator!\n");
			
//			Store result in magsweepdown
			magsweepup[i*nfield + k] = y_result[2];

//			Free the things we defined a while ago
			gsl_odeiv_evolve_free (e);
			gsl_odeiv_control_free (c);
			gsl_odeiv_step_free (s);
			free(param_list);
		}//end magsweepup for loop				
	}

	time(&stop);
//	printf("Took %d seconds to run\n", (int) stop-start);
//	free memory stuff
	gsl_matrix_free(NR);
	gsl_matrix_free(ND);
}

/*This function wraps Hsweep to unpack parameter values for the multi-threaded implementation
*/
void* Hsweep_wrapper(void* param_ptr) {
//	Unpack all of the parameters that have been packed in...
	double** params = (double**) param_ptr;
	int n = *(int*) params[0];
	int nfield = *(int*) params[1];
	double per_k2 = *(double*) params[2];
	double theta_offset = *(double*) params[3];
	double* hz_ptr = (double*) params[4];
	double* magsweepup = (double*) params[5];
	double* magsweepdown = (double*) params[6];
	double* j = (double*) params[7];
	double dx = *(double*) params[8];
	double dy = *(double*) params[9];
	double dz = *(double*) params[10];
	double Alpha = *(double*) params[11];
	double Ms = *(double*) params[12];
	double Mrs = *(double*) params[13];
	double K1 = *(double*) params[14];
	double hz_min = *(double*) params[15];
	double hz_max = *(double*) params[16];
	double j_min = *(double*) params[17];
	double j_max = *(double*) params[18];
	double t_f = *(double*) params[19];
	int thread_number = *(int*) params[20];
	int numthreads = *(int*) params[21];
	unsigned int j_timedep_funs = *(int*) params[22];
	unsigned int hz_timedep_funs = *(int*) params[23];
	
//	Call Hsweep
	Hsweep(n, nfield, per_k2, theta_offset, hz_ptr, magsweepup, magsweepdown, j, dx, dy, dz, Alpha, Ms, Mrs, K1, hz_min, hz_max, j_min, j_max, t_f, thread_number, numthreads, j_timedep_funs, hz_timedep_funs);
	
	return(NULL);
}

/*List of parameters to pass in
(1) K1
(2) Alpha
(3) Ms
(4) Mrs
(5) dx
(6) dy
(7) dz
(8) Min j
(9) Max j
(10) Min Hz
(11) Max Hz
(12) Integration time
(13) %K2
(14) thetaoff
(15) n
(16) nfield
(17) Number of threads
(18) Bit-wise OR of time-dependent current density functions to use
(19) Bit-wise OR of time-dependent external field functions to use
*/
int main(int argc, char ** argv) {
	//Check number of parameters on the command line
	if(argc != NUMARGS + 1) {
		printf("spinsim_gsl should be given %d parameters:\n", NUMARGS);
		printf("(1) K1\n(2) Alpha\n(3) Ms\n(4) Mrs\n(5) dx\n(6) dy\n(7) dz\n(8) Min j\n(9) Max j\n(10) Min Hz\n(11) Max Hz\n(12) Integration time\n(13) percent K2\n(14) thetaoff\n(15) n\n(16) nfield\n(17) Number of threads\n(18) Bit-wise OR of time-dependent current density functions to use\n(19) Bit-wise OR of time-dependent external field functions to use\n");
		exit(0);
	}
	//unpack and parse in all of the command-line arguments
	double K1 = atof(argv[1]); //unpack K1
	double Alpha = atof(argv[2]);
	double Ms = atof(argv[3]);
	double Mrs = atof(argv[4]);
	double dx = atof(argv[5]);
	double dy = atof(argv[6]);
	double dz = atof(argv[7]);
	double j_min = atof(argv[8]);
	double j_max = atof(argv[9]);
	double hz_min = atof(argv[10]);
	double hz_max = atof(argv[11]);
	double t_f = atof(argv[12]);
	double per_K2 = atof(argv[13]);
	double theta_offset = atof(argv[14]);
	int n = atoi(argv[15]);
	int n_field = atoi(argv[16]);
	int numthreads = atoi(argv[17]);
	unsigned int j_timedep_funs = atoi(argv[18]);
	unsigned int hz_timedep_funs = atoi(argv[19]);
	
	//for these values we'll only use the result from the 0'th threads calculations
	double * Hz = (double*) malloc(sizeof(double) * n_field);
	double * j = (double*) malloc(sizeof(double) * n);
	//Save pointers of "garbage variables" to be free'd later
	double** to_free_hz = (double**) malloc(sizeof(double*) * numthreads);
	double** to_free_j = (double**) malloc(sizeof(double*) * numthreads);
	//Note on threading: all threads will share this memory space. The threads are configured so they will
	//NOT crash into each other's data
	double * magsweepup = (double*) malloc(sizeof(double) * n * n_field);
	double * magsweepdown = (double*) malloc(sizeof(double) * n * n_field);

	//Keep track of the threads we've created
	int * thread_numbers = (int*) malloc(sizeof(int) * numthreads);
	//benchmarking
	time_t start, stop;
	//loop variables
	int count, count2;
	
	//Load in time-dependent field/current function parameters, if necessary
	if(j_timedep_funs > 1 || hz_timedep_funs > 1) {
		if(LoadTDParams() != 0) {
			printf("ERROR: LoadTDParams() returned an error. Exiting...");
			exit(-1);
		}
	}
	
	//Allocate pointer to all pthread_t 
	pthread_t *threads = (pthread_t *) malloc(numthreads * sizeof(*threads));
	//Attributes for creating threads
	pthread_attr_t pthread_custom_attr;
	pthread_attr_init(&pthread_custom_attr);
	//Specify that all threads are joinable
	pthread_attr_setdetachstate(&pthread_custom_attr,PTHREAD_CREATE_JOINABLE);
	
	time(&start);
	printf("Spinsim Hsweep Started!\n");

//	Single-threaded call
	if(numthreads == 1) {
		Hsweep(n, n_field, per_K2, theta_offset, Hz, magsweepup, magsweepdown, j, dx, dy, dz, Alpha, Ms, Mrs, K1, hz_min, hz_max, j_min, j_max, t_f, 0, numthreads, j_timedep_funs, hz_timedep_funs);
	}
//	Multi-threaded call, means packing and passing parameters!
	else {
//		Launch all of numthreads threads
		for(count = 0; count < numthreads; count++) {
//			Allocate space and pack in all of the parameters that will be passed to the count-th thread
			double** params = (double**) malloc(24 * sizeof(double *));
			params[0] = (double*) &n;
			params[1] = (double*) &n_field;
			params[2] = (double*) &per_K2;
			params[3] = (double*) &theta_offset;
			params[5] = (double*) magsweepup;
			params[6] = (double*) magsweepdown;
			//For j and Hz, we'll only save the results from the 0-th thread's calculations
			//as the arrays will be the same across all threads
			if(count == 0) {
				params[4] = (double*) Hz;
				params[7] = (double*) j;
			}
			else {
				params[4] = (double*) malloc(n_field * sizeof(double));
				params[7] = (double*) malloc(n * sizeof(double));
				to_free_hz[count-1] = (double*)params[4];
				to_free_j[count-1] = (double*)params[7];
			}
			//Keep packing parameters
			params[8] = (double*) &dx;
			params[9] = (double*) &dy;
			params[10] = (double*) &dz;
			params[11] = (double*) &Alpha;
			params[12] = (double*) &Ms;
			params[13] = (double*) &Mrs;
			params[14] = (double*) &K1;
			params[15] = (double*) &hz_min;
			params[16] = (double*) &hz_max;
			params[17] = (double*) &j_min;
			params[18] = (double*) &j_max;
			params[19] = (double*) &t_f;
			//save thread number, and how many threads there are
			thread_numbers[count] = count;
			params[20] = (double*) &thread_numbers[count];
			params[21] = (double*) &numthreads; //this is bad, we assume that double* and int* have same size
			params[22] = (double*) &j_timedep_funs;
			params[23] = (double*) &hz_timedep_funs;
			
			//call as an actual thread procedure
			pthread_create(&threads[count], &pthread_custom_attr, Hsweep_wrapper, (void*) params);
		}
	
		//Wait for all threads to finish (join)
		for (count=0; count < numthreads; count++)
		{
			pthread_join(threads[count],NULL);
		}

		//free garbage variables
		for(count = 0; count < numthreads-1; count++) {
			free(to_free_hz[count]);
			free(to_free_j[count]);
		}
	}
	
	time(&stop);
	printf("100 percent Done!");
	printf("Total time elapsed was %d seconds\n", (int) (stop-start));
	
	//output files for Hsweep output
	//Put together all of the file names based on the parameters given
	FILE* output_fileup, *output_filedown, *j_output, *hz_output;
	char filename[128];
	
	sprintf(filename, "magsweepup_output", n, n_field, per_K2, theta_offset);
	output_fileup = fopen(filename,"w");
	
	sprintf(filename, "magsweepdown_output", n, n_field, per_K2, theta_offset);
	output_filedown = fopen(filename,"w");
	
	sprintf(filename, "magsweep_output_j", n, n_field);
	j_output = fopen(filename,"w");
	
	sprintf(filename, "magsweep_output_hz", n, n_field);
	hz_output = fopen(filename,"w");
	
	//Write out magsweepdown data
	for(count = 0; count < n; count++) {
		for(count2 = 0; count2 < n_field; count2++) {
			fprintf(output_filedown, "%.3e  ", magsweepdown[count*n_field + count2]);
			
			if(magsweepdown[count*n_field + count2] > 0.0) {
				fprintf(output_filedown, " ");
			}
		}
		fprintf(output_filedown, "\n");
	}
	
	//write out magsweepup data
	for(count = 0; count < n; count++) {
		for(count2 = 0; count2 < n_field; count2++) {
			fprintf(output_fileup, "%.3e  ", magsweepup[count*n_field + count2]);
			
			if(magsweepup[count*n_field + count2] > 0.0) {
				fprintf(output_fileup, " ");
			}
		}
		fprintf(output_fileup, "\n");
	}
	
	//write out j and hz
	for(count = 0; count < n; count++) {
		fprintf(j_output, "%.3e\n", j[count]);
	}
	for(count = 0; count < n_field; count++) {
		fprintf(hz_output, "%.3e\n", Hz[count]);
	}
	
//	Close up all of our files
	fclose(output_fileup);
	fclose(output_filedown);
	fclose(j_output);
	fclose(hz_output);
	
	return(0);
}

